from __future__ import annotations

import argparse
import logging
import pathlib
import re
from collections.abc import Sequence
from typing import Any, Mapping, TypeVar, overload

import attr
import snakemake
from typing_extensions import override

from snakebids.exceptions import ConfigError, MisspecifiedCliFilterError
from snakebids.io.yaml import get_yaml_io
from snakebids.types import InputsConfig, OptionalFilter
from snakebids.utils.utils import to_resolved_path

# We define Path here in addition to pathlib to put both variables in globals()
# This way, users specifying a path type in their config.yaml can indicate
# either Path or pathlib.Path
Path = pathlib.Path

logger = logging.Logger(__name__)


class FilterParse(argparse.Action):
    """Class for parsing CLI filters in argparse."""

    # Constructor calling
    @override
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ):
        setattr(namespace, self.dest, {})
        if not values:
            return

        for pair in values:
            if "=" in pair:
                # split it into key and value
                key, value = pair.split("=", 1)
            elif ":" in pair:
                key, spec = pair.split(":", 1)
                spec = spec.lower()
                if spec == "optional":
                    value = OptionalFilter
                elif spec in ["required", "any"]:
                    value = True
                elif spec == "none":
                    value = False
                else:
                    # The flag isn't recognized
                    raise MisspecifiedCliFilterError(pair)
            else:
                raise MisspecifiedCliFilterError(pair)

            # assign into dictionary
            getattr(namespace, self.dest)[key] = value


class SnakemakeHelpAction(argparse.Action):
    """Class for printing snakemake usage in argparse."""

    @override
    def __call__(
        self,
        parser: argparse.ArgumentParser,
        namespace: argparse.Namespace,
        values: str | Sequence[Any] | None,
        option_string: str | None = None,
    ):
        snakemake.main(["-h"])  # type: ignore


@attr.frozen
class SnakebidsArgs:
    """Arguments for Snakebids App.

    Organizes the various options available for a generic Snakebids App, and store
    project specific arguments in a dict. Snakemake args are to be put in a list

    Attributes
    ----------
    force : bool
        Force output in a directory that already has contents
    outputdir : Path
        Directory to place outputs
    pybidsdb_dir : Path
        Directory to place pybids database
    snakemake_args : list of strings
        Arguments to pass on to Snakemake
    args_dict : dict[str, Any]
        Contains all the snakebids specific args. Meant to contain custom user args
        defined in config, as well as dynamic --filter-xx and --wildcard-xx args.
        These will eventually be printed in the new config.
    """

    force: bool
    outputdir: Path = attr.ib(converter=to_resolved_path)
    snakemake_args: list[str]
    args_dict: dict[str, Any]
    pybidsdb_dir: Path | None = None
    pybidsdb_reset: bool = False


def create_parser(include_snakemake: bool = False) -> argparse.ArgumentParser:
    """Generate basic Snakebids Parser.

    Includes the standard Snakebids arguments.
    """
    # The snakemake parser functionality did not seem to be implemented in the original
    # factoring. I left the logic here, but it should probably be removed if it's not
    # needed.
    if include_snakemake:
        # get snakemake parser
        smk_parser = snakemake.get_argument_parser()  # type: ignore

        # create parser
        parser = argparse.ArgumentParser(
            description="Snakebids helps build BIDS Apps with Snakemake",
            add_help=False,
            parents=[smk_parser],  # type: ignore
        )
    else:
        parser = argparse.ArgumentParser(
            description="Snakebids helps build BIDS Apps with Snakemake"
        )

    standard_group = parser.add_argument_group(
        "STANDARD", "Standard options for all snakebids apps"
    )

    standard_group.add_argument(
        "--workflow-mode",
        "--workflow_mode",
        "-W",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    # We use -x as the alias because both -f and -F are taken by snakemake
    standard_group.add_argument(
        "--force-conversion",
        "--force_conversion",
        "-x",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    standard_group.add_argument(
        "--pybidsdb-dir",
        "--pybidsdb_dir",
        action="store",
        help=(
            "Optional path to directory of SQLite databasefile for PyBIDS. "
            "If directory is passed and folder exists, indexing is skipped. "
            "If pybidsdb_reset is called, indexing will persist"
        ),
    )

    standard_group.add_argument(
        "--pybidsdb-reset",
        "--pybidsdb_reset",
        action="store_true",
        help=("Reindex existing PyBIDS SQLite database"),
    )

    # To be deprecated
    standard_group.add_argument(
        "--reset-db",
        "--reset_db",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    standard_group.add_argument(
        "--force-output",
        "--force_output",
        action="store_true",
        help="Force output in a new directory that already has contents",
    )

    standard_group.add_argument(
        "--retrofit",
        action="store_true",
        help=argparse.SUPPRESS,
    )

    # add option for printing out snakemake usage
    standard_group.add_argument(
        "--help-snakemake",
        "--help_snakemake",
        nargs=0,
        action=SnakemakeHelpAction,
        help=(
            "Options to Snakemake can also be passed directly at the "
            "command-line, use this to print Snakemake usage"
        ),
    )
    return parser


def _find_type(name: str, *, yamlsafe: bool = True) -> type[Any]:
    import importlib

    if name == "Path":
        return Path
    *module_name, obj_name = name.split(".") if "." in name else ("builtins", name)
    try:
        type_ = getattr(importlib.import_module(".".join(module_name)), obj_name)
    except (ImportError, AttributeError) as err:
        msg = f"{name} could not be resolved"
        raise ConfigError(msg) from err
    if not callable(type_):
        msg = f"{name} cannot be used as a type"
        raise ConfigError(msg)
    if yamlsafe and type_ not in get_yaml_io().representer.yaml_representers:
        msg = f"{name} cannot be serialized into yaml"
        raise ConfigError(msg)
    return type_


def add_dynamic_args(
    parser: argparse.ArgumentParser,
    parse_args: Mapping[str, Any],
    pybids_inputs: InputsConfig,
) -> None:
    """Add --filter-<comp> and --wildcards-<comp> arguments to CLI."""
    # create parser group for app options
    app_group = parser.add_argument_group("SNAKEBIDS", "Options for snakebids app")

    # update the parser with config options
    for name, arg in parse_args.items():
        # Convert type annotations from strings to class types
        # We first check that the type annotation is, in fact,
        # a str to allow the edge case where it's already
        # been converted
        if "type" in arg:
            arg_dict = {**arg, "type": _find_type(str(arg["type"]))}
        else:
            arg_dict = arg
        app_group.add_argument(
            *_make_underscore_dash_aliases(name),
            **arg_dict,  # type: ignore
        )

    # general parser for
    # --filter_{input_type} {key1}={value1} {key2}={value2}...
    # create filter parsers, one for each input_type
    filter_opts = parser.add_argument_group(
        "BIDS FILTERS",
        "Filters to customize PyBIDS get() as key=value pairs, or as "
        "key:{REQUIRED|OPTIONAL|NONE} (case-insensitive), to enforce the presence or "
        "absence of values for that key.",
    )

    for input_type in pybids_inputs:
        argnames = (f"--filter-{input_type}", f"--filter_{input_type}")
        filters = pybids_inputs[input_type].get("filters", {})
        arglist_default = [f"{key}={value}" for (key, value) in filters.items()]

        filter_opts.add_argument(
            *argnames,
            nargs="+",
            action=FilterParse,
            help=f"(default: {' '.join(arglist_default)})",
        )

    # general parser for
    # --wildcards_{input_type} {wildcard1} {wildcard2} ...
    # create wildcards parsers, one for each input_type
    wildcards_opts = parser.add_argument_group(
        "INPUT WILDCARDS",
        "File path entities to use as wildcards in snakemake",
    )

    for input_type in pybids_inputs:
        argnames = (f"--wildcards-{input_type}", f"--wildcards_{input_type}")
        arglist_default = [
            f"{wc}" for wc in pybids_inputs[input_type].get("wildcards", [])
        ]

        wildcards_opts.add_argument(
            *argnames,
            nargs="+",
            help=f"(default: {' '.join(arglist_default)})",
        )

    override_opts = parser.add_argument_group(
        "PATH OVERRIDE",
        (
            "Options for overriding BIDS by specifying absolute paths "
            "that include wildcards, e.g.: "
            "/path/to/my_data/{subject}/t1.nii.gz"
        ),
    )

    # create path override parser
    for input_type in pybids_inputs:
        argnames = (f"--path-{input_type}", f"--path_{input_type}")
        override_opts.add_argument(*argnames, default=None)


def parse_snakebids_args(parser: argparse.ArgumentParser) -> SnakebidsArgs:
    """Parse built-in snakebids arguments."""
    all_args = parser.parse_known_args()
    if all_args[0].workflow_mode:
        logger.warning(
            "--workflow-mode is deprecated, and no longer has any effect. Workflow "
            "mode is automatially activated when choosing the snakemake root "
            "directory, or a subfolder of root/results, as the output directory."
        )
    if all_args[0].retrofit:
        logger.warning(
            "--retrofit is deprecated and no longer has any effect. To run this "
            "snakebids app on an old output (e.g. if snakebids raises an error because "
            "the directory already has contents, use the --force-output flag."
        )
    if all_args[0].force_conversion:
        logger.warning("--force-conversion is deprecated and no longer has any effect.")
    if all_args[0].reset_db:
        logger.warning(
            "--reset-db/--reset_db will be deprecated in a future release. To reset "
            "the pybids database, use the new --pybidsdb-reset flag instead."
        )
    return SnakebidsArgs(
        snakemake_args=all_args[1],
        # resolve all path items to get absolute paths
        args_dict={k: _resolve_path(v) for k, v in all_args[0].__dict__.items()},
        force=all_args[0].force_output,
        outputdir=Path(all_args[0].output_dir).resolve(),
        pybidsdb_dir=(
            None
            if all_args[0].pybidsdb_dir is None
            else Path(all_args[0].pybidsdb_dir).resolve()
        ),
        pybidsdb_reset=all_args[0].pybidsdb_reset or all_args[0].reset_db,
    )


def _make_underscore_dash_aliases(name: str) -> set[str]:
    """Generate --dash-arg aliases for --dash_args and vice versa.

    If no dashes or underscores are in the argument name, a tuple containing just the
    original arg name will be returned

    Parameters
    ----------
    name : str
        Argument to generate conversion for

    Returns
    -------
    set of strings
        Converted args
    """
    if match := re.match(r"^--(.+)$", name):
        name_part = match.group(1)
        return {
            name,
            re.sub(r"\_", "-", name),
            f"--{re.sub(r'-', '_', name_part)}",
        }
    return {name}


_T = TypeVar("_T")


@overload
def _resolve_path(path_candidate: Sequence[Any]) -> list[Any]:
    ...


@overload
def _resolve_path(path_candidate: _T) -> _T:
    ...


def _resolve_path(path_candidate: Any) -> Any:
    """Resolve paths or list of paths, or return argument unchanged.

    Parameters
    ----------
    command : list, os.Pathlike, object
        command to run

    Returns
    -------
        If os.Pathlike or list  of os.Pathlike, the same paths resolved.
        Otherwise, the argument unchanged.
    """
    if isinstance(path_candidate, Sequence) and not isinstance(path_candidate, str):
        return [
            _resolve_path(p)  # type: ignore[reportUnknownArgumentType]
            for p in path_candidate  # type: ignore[reportUnknownVariableType]
        ]

    if isinstance(path_candidate, Path):
        return Path(path_candidate).resolve()

    return path_candidate



from typing import Any, Callable, Generic, TypeVar

T = TypeVar("T")

class UserProperty(Generic[T]):
    def __init__(self, method: Callable[[Any], T], /) -> None: ...
    def __get__(self, obj: Any, objtype: type[Any] = ...) -> T: ...



import json
import logging
import subprocess as sp
import tempfile

import attr

from snakebids.app import SnakeBidsApp
from snakebids.exceptions import SnakebidsPluginError

_logger = logging.getLogger(__name__)


class InvalidBidsError(SnakebidsPluginError):
    """Error raised if an input BIDS dataset is invalid."""


@attr.define
class BidsValidator:
    """Perform validation of a BIDS dataset using the bids-validator.

    If the dataset is not valid according to the BIDS specifications, an
    InvalidBidsError is raised.

    Parameters
    ----------
    raise_invalid_bids : bool
        Flag to indicate whether InvalidBidsError should be raised if BIDS
        validation fails. Default to True.

    """

    raise_invalid_bids: bool = attr.field(default=True)

    def __call__(self, app: SnakeBidsApp) -> None:
        """Perform BIDS validation of dataset.

        Parameters
        ----------
        app
            Snakebids application to be run

        Raises
        ------
        InvalidBidsError
            Raised when the input BIDS directory does not pass validation with
            the bids-validator
        """
        # Skip bids validation
        if app.config["plugins.validator.skip"]:
            return

        validator_config_dict = {"ignoredFiles": ["/participants.tsv"]}

        with tempfile.NamedTemporaryFile(mode="w+", suffix=".json") as temp:
            temp.write(json.dumps(validator_config_dict))
            temp.flush()
            try:
                sp.check_call(
                    ["bids-validator", app.config["bids_dir"], "-c", temp.name]
                )

                # If successfully bids-validation performed
                app.config["plugins.validator.success"] = True
            except FileNotFoundError:
                # If the bids-validator call can't be made
                app.config["plugins.validator.success"] = False
                _logger.warning(
                    "Missing bids-validator installation - falling back to pybids "
                    "validation."
                )
            # Any other bids-validator error
            except sp.CalledProcessError as err:
                app.config["plugins.validator.success"] = False
                if self.raise_invalid_bids:
                    raise InvalidBidsError from err



"""Tools to generate a Snakemake-based BIDS app."""
from __future__ import annotations

import argparse
import logging
import sys
from importlib import metadata as impm
from os import PathLike
from pathlib import Path
from typing import Any, Callable

import attr
import boutiques.creator as bc  # type: ignore
import snakemake
from snakemake.io import load_configfile

from snakebids.cli import (
    SnakebidsArgs,
    add_dynamic_args,
    create_parser,
    parse_snakebids_args,
)
from snakebids.exceptions import ConfigError, RunError
from snakebids.io.config import write_config
from snakebids.types import OptionalFilter
from snakebids.utils.output import (
    prepare_bidsapp_output,
    write_output_mode,
)
from snakebids.utils.utils import DEPRECATION_FLAG, to_resolved_path

logger = logging.Logger(__name__)


SNAKEFILE_CHOICES = [
    "Snakefile",
    "snakefile",
    "workflow/Snakefile",
    "workflow/snakefile",
]


CONFIGFILE_CHOICES = [
    "config/snakebids.yml",
    "config/snakebids.yaml",
    "config/snakebids.json",
    "snakebids.yml",
    "snakebids.yaml",
    "snakebids.json",
    "config.yml",
    "config.yaml",
    "config.json",
    "config/config.json",
    "config/config.yml",
    "config/config.yaml",
]


def _get_file_paths(
    choices: list[str], file_name: str
) -> Callable[[SnakeBidsApp], Path]:
    def wrapper(self: SnakeBidsApp):
        for path in choices:
            if (self.snakemake_dir / path).exists():
                if file_name == "config":
                    return Path(path)
                # else, snakefile
                return Path(self.snakemake_dir, path)

        msg = f"Error: no {file_name} file found, tried {', '.join(choices)}."
        raise ConfigError(msg)

    return wrapper


def _get_app_version(self: SnakeBidsApp) -> str | None:
    """Attempt to get the app version, returning None if we can't.

    This will succeed only if the following conditions are true:

    1. The Snakebids app is a distribution package installed in the current
       environment.
    2. The app's distribution package has the same name as this
       SnakeBidsApp's snakemake_dir
    """
    try:
        return impm.version(self.snakemake_dir.name)
    except impm.PackageNotFoundError:
        logger.warning(
            "App version not found; will be recorded in output as 'unknown'. "
            "If this is unexpected, please contact the app maintainer."
        )
        return None


@attr.define(slots=False)
class SnakeBidsApp:
    """Snakebids app with config and arguments.

    Parameters
    ----------
    snakemake_dir : str | Path
        Root directory of the snakebids app, containing the config file and workflow
        files.
    parser
        Parser including only the arguments specific to this Snakebids app, as specified
        in the config file. By default, it will use `create_parser()` from `cli.py`
    configfile_path
        Relative path to config file (relative to snakemake_dir). By default,
        autocalculates based on snamake_dir
    snakefile_path
        Absolute path to the input Snakefile. By default, autocalculates based on
        snakemake_dir::

            join(snakemake_dir, snakefile_path)
    config
        Contains all the configuration variables parsed from the config file
        and generated during the initialization of the SnakeBidsApp.
    args
        Arguments to use when running the app. By default, generated using the parser
        attribute, autopopulated with args from `config.py`
    plugins
        List of methods to be called after CLI parsing.

        Each callable in ``plugins`` should take, as a single argument, a reference to
        the ``SnakeBidsApp``. Plugins may perform any arbitrary side effects, including
        updates to the config dictionary, validation of inputs, optimization, or other
        enhancements to the snakebids app.

        CLI parameters may be read from ``SnakeBidsApp.config``. Plugins are responsible
        for documenting what properties they expect to find in the config.

        Every plugin should return either:

        - Nothing, in which case any changes to the SnakeBidsApp will persist in the
          workflow.
        - A ``SnakeBidsApp``, which will replace the existing instance, so this option
          should be used with care.

    """

    snakemake_dir: Path = attr.ib(converter=to_resolved_path)
    plugins: list[Callable[[SnakeBidsApp], None | SnakeBidsApp]] = attr.Factory(list)
    skip_parse_args: bool = False
    parser: argparse.ArgumentParser = attr.Factory(create_parser)
    configfile_path: Path = attr.Factory(
        _get_file_paths(CONFIGFILE_CHOICES, "config"), takes_self=True
    )
    snakefile_path: Path = attr.Factory(
        _get_file_paths(SNAKEFILE_CHOICES, "Snakefile"), takes_self=True
    )
    config: dict[str, Any] = attr.Factory(
        lambda self: load_configfile(self.snakemake_dir / self.configfile_path),
        takes_self=True,
    )
    version: str | None = attr.Factory(_get_app_version, takes_self=True)
    args: SnakebidsArgs | None = None

    def run_snakemake(self) -> None:
        """Run snakemake with the given config, after applying plugins."""
        # If no SnakebidsArgs were provided on class instantiation, we compute args
        # using the provided parser
        if not self.args:
            # Dynamic args include --filter-... and --wildcards-... . They depend on the
            # config
            add_dynamic_args(
                self.parser, self.config["parse_args"], self.config["pybids_inputs"]
            )
            self.args = parse_snakebids_args(self.parser)

        # Update our config file:
        # - Add path to snakefile to the config so workflows can grab files relative to
        #    the snakefile folder
        # - Add info from args
        # - Set mode (bidsapp or workflow) and output_dir appropriately
        update_config(self.config, self.args)

        self.config["snakemake_dir"] = self.snakemake_dir
        self.config["snakefile"] = self.snakefile_path

        # Update config with pybids settings
        self.config["pybidsdb_dir"] = self.args.pybidsdb_dir
        self.config["pybidsdb_reset"] = self.args.pybidsdb_reset
        self.config[
            "pybids_db_dir"
        ] = f"{DEPRECATION_FLAG}{self.args.pybidsdb_dir}{DEPRECATION_FLAG}"
        self.config[
            "pybids_db_reset"
        ] = f"{DEPRECATION_FLAG}{int(self.args.pybidsdb_reset)}{DEPRECATION_FLAG}"

        # First, handle outputs in snakebids_root or results folder
        try:
            # py3.9 has the Path.is_relative() function. But as long as we support py38
            # and lower, this is the easiest way
            self.args.outputdir.resolve().relative_to(self.snakemake_dir / "results")
            relative_to_results = True
        except ValueError:
            relative_to_results = False

        if self.snakemake_dir == self.args.outputdir.resolve() or relative_to_results:
            write_output_mode(self.snakemake_dir / ".snakebids", "workflow")

            new_config_file = self.snakemake_dir / self.configfile_path
            cwd = self.snakemake_dir

            if self.config["output_dir"] == self.snakemake_dir.resolve():
                self.config["output_dir"] /= "results"
                self.config["root"] = "results"
                # Print a friendly warning that the output directory will change
                logger.info(
                    "You specified your output to be in the snakebids directory, so "
                    "we're automatically putting your outputs in the results "
                    "subdirectory.\nYou'll find your results in `%s`",
                    (self.snakemake_dir / "results").resolve(),
                )
            else:
                self.config["root"] = ""

        # else, we run in bidsapp mode
        else:
            # Attempt to prepare the output folder. Anything going wrong will raise a
            # RunError, as described in the docstring
            try:
                prepare_bidsapp_output(self.args.outputdir, self.args.force)
            except RunError as err:
                print(err.msg)
                sys.exit(1)
            cwd = self.args.outputdir
            new_config_file = self.args.outputdir / self.configfile_path
            self.config["root"] = ""

        app = self
        for plugin in self.plugins:
            app = plugin(app) or app

        # Write the config file
        write_config(
            config_file=new_config_file,
            data=dict(
                app.config,
                snakemake_version=impm.version("snakemake"),
                snakebids_version=impm.version("snakebids"),
                app_version=app.version or "unknown",
            ),
            force_overwrite=True,
        )

        # Run snakemake (passing any leftover args from argparse)
        # Filter any blank strings before submitting
        snakemake.main(  # type: ignore
            [
                *filter(
                    None,
                    [
                        *app.config["targets_by_analysis_level"][
                            app.config["analysis_level"]
                        ],
                        "--snakefile",
                        str(app.snakefile_path),
                        "--directory",
                        str(cwd),
                        "--configfile",
                        str(new_config_file.resolve()),
                        *app.config["snakemake_args"],
                    ],
                )
            ]
        )

    def create_descriptor(self, out_file: PathLike[str] | str) -> None:
        """Generate a boutiques descriptor for this Snakebids app."""
        new_descriptor = bc.CreateDescriptor(  # type: ignore
            self.parser, execname="run.py"
        )
        new_descriptor.save(out_file)  # type: ignore


def update_config(config: dict[str, Any], snakebids_args: SnakebidsArgs) -> None:
    """Add snakebids arguments to config in-place."""
    config.update({"snakemake_args": snakebids_args.snakemake_args})

    # argparse adds filter_{input_type}
    # we want to update the pybids_inputs dict with this, then remove the
    # filter_{input_type} dict
    pybids_inputs = config["pybids_inputs"]
    args = snakebids_args.args_dict
    for input_type in pybids_inputs:
        arg_filter_dict = args[f"filter_{input_type}"]
        if arg_filter_dict is not None:
            pybids_inputs[input_type].setdefault("filters", {})
            for entity, filter_ in arg_filter_dict.items():
                if filter_ is OptionalFilter:
                    pybids_inputs[input_type]["filters"].pop(entity, None)
                else:
                    pybids_inputs[input_type]["filters"][entity] = filter_
        del args[f"filter_{input_type}"]

    # add cmdline defined wildcards from the list:
    # wildcards_{input_type}
    for input_type in pybids_inputs:
        wildcards_list = args[f"wildcards_{input_type}"]
        if wildcards_list is not None:
            pybids_inputs[input_type].setdefault("wildcards", [])
            pybids_inputs[input_type]["wildcards"] += wildcards_list
        del args[f"wildcards_{input_type}"]

    # add custom input paths to
    # config['pybids_inputs'][input_type]['custom_path']
    for input_type in pybids_inputs:
        custom_path = args[f"path_{input_type}"]
        if custom_path is not None:
            pybids_inputs[input_type]["custom_path"] = Path(custom_path).resolve()
        del args[f"path_{input_type}"]

    # add snakebids arguments to config
    config.update(args)

    # replace paths with realpaths
    config["bids_dir"] = Path(config["bids_dir"]).resolve()
    config["output_dir"] = Path(config["output_dir"]).resolve()
